from typing import List, Dict

from confluent_kafka import TIMESTAMP_NOT_AVAILABLE

from ..service import Service
from ..logger import init_logger
from ..topics import MACHINE_EVENTS, PROPOSAL_TOPIC
from ..messages.message import Headers
from ..config import KAFKA_SERVER_URL
from ..messages.update_message_types import SetMachineMessage, UpdateMessage
from ..event_utls.consumer import consume

_logger = init_logger(__name__)


class MachineService(Service):
    def __init__(self, name, write, read):
        super().__init__(name)
        self.write = write
        self.read = read
        self.updated_devices: Dict[int, List[List[str]]] = {}

    def simulate(self):
        pass

    def post_init(self):
        for msg in consume([PROPOSAL_TOPIC], timeout_poll=0.1, group_id=self.type, bootstrap_server=KAFKA_SERVER_URL):
            self.machine_input_handler(msg)

    def machine_input_handler(self, msg, **kwargs):
        timestamp_type, timestamp = msg.timestamp()
        if timestamp_type == TIMESTAMP_NOT_AVAILABLE:
            _logger.debug(f"[{self.name}] receive a message without a timestamp")
            return
        headers = Headers.from_kafka_headers(msg.headers())
        received_package_id = headers.package_id
        _logger.debug(f'[{self.name}] call machine_input_handler receive headers: {str(headers)}')

        message = SetMachineMessage.deserialize([msg])

        _updated_devices = []
        for write_command in message.write_commands:
            self.write(write_command.channel, write_command.device_name, write_command.property, **write_command.params)
            _updated_devices.append(f"{write_command.channel}/{write_command.device_name}")

        self.updated_event(package_id=received_package_id,
                           msg=UpdateMessage(source=headers.msg_type, updated=_updated_devices))

        _logger.debug(f"[{self.name}] end machine_input_handler")

    def updated_event(self, package_id, msg):
        self.producer.sync_produce(MACHINE_EVENTS, msg.serialize(), Headers(package_id=package_id, source=self.type))



from typing import List, Dict

from confluent_kafka import TIMESTAMP_NOT_AVAILABLE

from ..service import Service
from ..logger import init_logger
from ..topics import MACHINE_EVENTS, SIMULATION_EVENTS
from ..messages.message import Headers
from ..config import KAFKA_SERVER_URL
from ..messages.update_message_types import UpdateMessage
from ..event_utls.consumer import consume

_logger = init_logger(__name__)


class SimulationService(Service):
    def __init__(self, name, write, read, channels: List = []):
        super().__init__(name)
        self.category = 'Simulation'
        self.read = read

        self._write = write
        self._updated_channels = []

        self.updated_devices: Dict[int, List[List[str]]] = {}
        self.channels = set(channels)
        self.__stop_consumer_loop = False

    def write(self, channel: str, device: str, _property, **kwargs):
        self._updated_channels.append(channel)
        self._write(channel, device, _property, **kwargs)

    def _is_message_for_me(self, msg: UpdateMessage):
        if len(self.channels) == 0:
            return True
        for channel in msg.updated:
            if channel in self.channels:
                return True
        return False

    def simulate(self):
        pass

    def stop(self):
        self.__stop_consumer_loop = True
        super().stop()

    def __consumer_stop_hook(self):
        return self.__stop_consumer_loop

    def post_init(self):
        for msg in consume([MACHINE_EVENTS], stop_hook=self.__consumer_stop_hook, timeout_poll=0.1, group_id=self.type, bootstrap_server=KAFKA_SERVER_URL):
            self.machine_input_handler(msg)

    def machine_input_handler(self, msg, **kwargs):
        timestamp_type, timestamp = msg.timestamp()
        if timestamp_type == TIMESTAMP_NOT_AVAILABLE:
            _logger.debug(f"[{self.name}] receive a message without a timestamp")
            return
        headers = Headers.from_kafka_headers(msg.headers())
        received_package_id = headers.package_id
        _logger.debug(f'[{self.name}] call machine_input_handler receive headers: {str(headers)}')

        message = UpdateMessage.deserialize([msg])

        if self._is_message_for_me(message):
            _logger.debug("simulate...")
            self.simulate()
            self.sim_updated_event(package_id=received_package_id,
                                   msg=UpdateMessage(source=headers.msg_type, updated=self._updated_channels))
            self._updated_channels = []

        _logger.debug(f"[{self.name}] end machine_input_handler")

    def sim_updated_event(self, package_id, msg):
        self.producer.sync_produce(SIMULATION_EVENTS, msg.serialize(), Headers(package_id=package_id, source=self.type))



import threading

from ..logger import logging
from .consumer import consume

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


class ConsumerThread(threading.Thread):
    def __init__(self, topics, group_id, bootstrap_server, message_handler, consumer_started_hook, **kwargs):
        super(ConsumerThread, self).__init__()
        self.lock = threading.Lock()
        self.bootstrap_server = bootstrap_server
        self.name = group_id
        #        self.daemon = True
        self._stop_event = threading.Event()
        self._ready_event = threading.Event()
        self.message_handlers = [message_handler] if message_handler else []
        self.consumer_started_hook = consumer_started_hook if consumer_started_hook else None
        self.topics = topics

        self.group_id = group_id
        self.kwargs = kwargs

    def stop(self):
        self._stop_event.set()

    def wait_until_ready(self):
        self._ready_event.wait()

    def restart(self):
        self._stop_event.clear()

    def stopped(self):
        return self._stop_event.is_set()

    def subscribe(self, msg_handler):
        with self.lock:
            self.message_handlers.append(msg_handler)

    def unsubscribe(self, msg_handler):
        func_id = id(msg_handler)
        for idx, active_msg_handler in enumerate(self.message_handlers):
            if func_id == id(active_msg_handler):
                with self.lock:
                    self.message_handlers.pop(idx)
                return

    def run(self):
        def thread_started(msg):
            if not self._ready_event.is_set():
                self._ready_event.set()

        for msg in consume(topics=self.topics, timeout_poll=0.1, group_id=self.group_id, consumer_started_hook=self.consumer_started_hook, all_msg_hook=thread_started, stop_hook=self.stopped, bootstrap_server=self.bootstrap_server):
            with self.lock:
                for message_handler in self.message_handlers:
                    message_handler(msg, **self.kwargs)



import http

from flask import Flask, request, make_response
from src.acss_core.logger import init_logger
from src.console.compose_client import DockerComposeClient
from src.acss_core.client.kafka_client import KafkaPipeClient


_logger = init_logger(__name__)


def _build_cors_prelight_response():
    response = make_response()
    response.headers.add("Access-Control-Allow-Origin", "*")
    response.headers.add("Access-Control-Allow-Headers", "*")
    response.headers.add("Access-Control-Allow-Methods", "*")
    return response


def create_app():
    app = Flask(__name__)
    app.config['JSON_AS_ASCII'] = False
    return app


app = create_app()
pipe_client = KafkaPipeClient()

global compose_client
compose_client = DockerComposeClient(64)


@app.route('/')
def is_alive():
    return "I'm alive\n"


@app.route('/start/services', methods=['GET', 'OPTIONS'])
def get_service():
    if request.method == "OPTIONS":  # CORS preflight
        return _build_cors_prelight_response()
    else:
        services = compose_client.get_services()
        print(services)
        return {'services': services}, http.HTTPStatus.OK, {'Content-Type': 'application/json; charset=utf-8', "Access-Control-Allow-Origin": "*"}


@app.route('/start/<name>/<service_type>/<mode>', methods=['GET', 'OPTIONS'])
def start_service(name, service_type, mode):
    if request.method == "OPTIONS":  # CORS preflight
        return _build_cors_prelight_response()
    else:
        services = compose_client.get_services()
        print(services)
        if service_type not in services:
            return f"service type '{service_type}' does not exists.", http.HTTPStatus.BAD_REQUEST, {'Content-Type': 'application/json; charset=utf-8', "Access-Control-Allow-Origin": "*"}

        compose_client.run_service(name=name, service_type=service_type, mode=mode)
        return '', http.HTTPStatus.OK, {'Content-Type': 'application/json; charset=utf-8', "Access-Control-Allow-Origin": "*"}


@app.route('/stop/<name>', methods=['GET', 'OPTIONS'])
def stop_service(name):
    if request.method == "OPTIONS":  # CORS preflight
        return _build_cors_prelight_response()
    else:
        pipe_client.stop_service(name)

        return '', http.HTTPStatus.OK, {'Content-Type': 'application/json; charset=utf-8', "Access-Control-Allow-Origin": "*"}


@app.route('/services', methods=['GET', 'OPTIONS'])
def get_running_services():
    if request.method == "OPTIONS":  # CORS preflight
        return _build_cors_prelight_response()
    else:
        res = pipe_client.get_running_service_info()
        return {'services': res}, http.HTTPStatus.OK, {'Content-Type': 'application/json; charset=utf-8', "Access-Control-Allow-Origin": "*", "Access-Control-Allow-Headers": "*", "Access-Control-Allow-Methods": "*"}


if __name__ == '__main__':
    app.run(debug=True, port=5002, host='0.0.0.0')



import json
from typing import List
from collections import namedtuple
import time
import requests

from ..topics import CONTROL_TOPIC
from ..config import KAFKA_SERVER_URL, OBSERVER_URL, REGISTER_URL
from ..topics import RECONFIG_TOPIC, SERVICE_INPUT_TOPIC
from ..messages.simple_service_message import SimpleServiceMessage
from ..messages.agent_result_message import AgentResultMessage
from ..messages.message import Headers, ControlMessage
from ..event_utls.producer import Producer
from ..utils.utils import generate_unique_id, wait_until_server_is_online
from ..logger import logging


_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)

ServiceId = namedtuple('ServiceId', ['id', 'name'])


class KafkaPipeClient():
    def __init__(self, kafka_broker_url: str = KAFKA_SERVER_URL) -> None:
        self.kafka_broker_url = kafka_broker_url
        self.cc_producer = Producer(kafka_broker_url)
        wait_until_server_is_online(REGISTER_URL, _logger)
        wait_until_server_is_online(OBSERVER_URL, _logger)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        pass

    def wait_for_services(self, names: List[str], timeout: float = 30, sleep_time: float = 0.1, option='start', status=None):
        expected_len = 0 if option == 'shutdown' else len(names)

        t = 0.0
        names_to_be_found = set(names)
        while(t < timeout):
            name_type_set = self.get_service_type_and_name_set(status=status)
            if len(names_to_be_found.intersection(name_type_set)) == expected_len:
                return True
            time.sleep(sleep_time)
            t += sleep_time
        return False

    def run_service(self, service_type, params=None):

        package_id = generate_unique_id()

        self.cc_producer.sync_produce(topic=SERVICE_INPUT_TOPIC,
                                      value=SimpleServiceMessage(params=params).serialize(),
                                      headers=Headers(package_id=package_id, source='console', only_for=[service_type]))

        return ServiceId(package_id, service_type)

    def stop_service(self, name):
        package_id = generate_unique_id()
        msg = ControlMessage(name=name, command='stop')

        self.cc_producer.sync_produce(topic=CONTROL_TOPIC,
                                      value=msg.serialize(),
                                      headers=Headers(package_id=package_id, source='console', only_for=[name]))

    def reconfig(self, names: List[str], config_data: dict):
        package_id = generate_unique_id()
        self.cc_producer.all_partitions_produce(RECONFIG_TOPIC,
                                                value=json.dumps(config_data),
                                                headers=Headers(package_id=package_id,  source='console', only_for=names,
                                                                msg_type=None))
        return package_id

    def get_running_service_info(self):
        res = requests.get(f'http://{REGISTER_URL}/services/')
        if res.status_code == 200:
            return res.json()['services']
        _logger.error(f"receive error code {res.status_code}.")
        return None

    def get_service_type_and_name_set(self, status=None):
        type_name_hash = set()
        for name, info in self.get_running_service_info().items():
            print(name, info['type'], info['status'])
            if status is None or info['status'] == status:
                type_name_hash.add(name)
                type_name_hash.add(info['type'])
        return type_name_hash

    def get_running_services(self):
        return list(self.get_running_service_info().keys())

    def _get_results_from_observer(self, package_ids: List[str], service_type: str, timeout=30, poll_time=0.05):
        ids = ','.join(package_ids)
        res = requests.get(f'http://{OBSERVER_URL}/find/{ids}/{service_type}')
        if res.status_code == 200:
            return res.json()
        time_counter = 0.0
        while(res.status_code == 202):
            time.sleep(poll_time)
            time_counter += poll_time
            res = requests.get(f'http://{OBSERVER_URL}/find/{ids}/{service_type}')
            if res.status_code == 200:
                _logger.debug(f"service name = {service_type}")
                return res.json()
            if time_counter >= timeout:
                _logger.error(f"timeout reached for package_id={ids} and service_type={service_type}.")
                return None

        _logger.error(f"receive error code {res.status_code}.")
        return None

    def wait_for_simulation(self, service_id: ServiceId, sim_type: str):
        return self._get_results_from_observer([service_id.id], sim_type, timeout=30, poll_time=0.05)

    def get_service_results(self, service_id: ServiceId):
        agt_result = self._get_results_from_observer([service_id.id], service_id.name, timeout=30, poll_time=0.05)
        return [AgentResultMessage.from_byte_string(agt) for agt in agt_result['data']]



import yaml
import logging
import os

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


PATH_TO_ACSS_SERVICES_ROOT = os.environ.get('PATH_TO_ACSS_SERVICES_ROOT')

ACSS_CONFIG_FILEPATH = os.environ.get('ACSS_CONFIG_FILEPATH', 'ml-pipe-config.yaml')

with open(ACSS_CONFIG_FILEPATH, 'r') as file:
    pipe_conf = yaml.load(file, Loader=yaml.FullLoader)

msg_bus_settings = pipe_conf.get('msg_bus')
if msg_bus_settings is None:
    raise ValueError(f"Message bus is not configured. Please add msg_bus key to config file.")
else:
    broker_urls = msg_bus_settings.get("broker_urls")
    if broker_urls != None:
        brokers = [b.strip() for b in broker_urls.split(',')]
        KAFKA_SERVER_URL = brokers[0]
    if KAFKA_SERVER_URL == None:
        raise ValueError(f"No broker is set.")

sim_settings = pipe_conf.get('simulation')
if sim_settings is None:
    raise ValueError(f"Simulation is not configured. Please add simulation key to config file.")
MYSQL_USER = sim_settings.get("sim_db_usr")
if MYSQL_USER == None:
    raise ValueError("sim_db_usr key is not set in config.")
MYSQL_PW = sim_settings.get("sim_db_pw")
if MYSQL_PW == None:
    raise ValueError("sim_db_pw key is not set in config.")
MYSQL_HOST, MYSQL_PORT = sim_settings.get('sim_db_url').strip().split(':')
if MYSQL_HOST == None:
    raise ValueError("sim_db_url key is not set in config.")

obs_settings = pipe_conf.get("observer")
if obs_settings is None:
    raise ValueError(f"Observer is not configured. Please add observer key to config file.")
OBSERVER_URL = obs_settings.get('url')
if OBSERVER_URL is None:
    raise ValueError("url key for observer frontend is not set in config.")
PIPE_EVENT_DB_USER = obs_settings.get("event_db_usr")
if PIPE_EVENT_DB_USER == None:
    raise ValueError("event_db_usr key is not set in config.")
PIPE_EVENT_DB_PW = obs_settings.get("event_db_pw")
if PIPE_EVENT_DB_PW == None:
    raise ValueError("event_db_pw key is not set in config.")
PIPE_EVENT_DB_URL = obs_settings.get("event_db_url", "observer_database")

register_settings = pipe_conf.get("register")
if obs_settings is None:
    raise ValueError(f"Register is not configured. Please add register key to config file.")
REGISTER_URL = register_settings.get('url')

if REGISTER_URL is None:
    raise ValueError("url key for register frontend is not set in config.")


LOGSTASH_SERVER_URL = None
LOGSTASH_PORT = None
logging_settings = pipe_conf.get("logging")
if logging_settings is not None:
    logstash_url = logging_settings.get("logstash_url")
    if logstash_url is not None:
        LOGSTASH_SERVER_URL, LOGSTASH_PORT = logstash_url.strip().split(':')

_logger.info(f"KAFKA_SERVER_URL={KAFKA_SERVER_URL}")
_logger.info(f"MYSQL_HOST={MYSQL_HOST}")
_logger.info(f"MYSQL_PORT={MYSQL_PORT}")
_logger.info(f"MYSQL_USER={MYSQL_USER}")



import time
import requests
from acss_core.config import OBSERVER_URL
from acss_core.simulation.MachineService import MachineService
from ..messages.update_message_types import WriteCommand, SetMachineMessage
from ..topics import PROPOSAL_TOPIC
from ..messages.message import Headers
from ..logger import init_logger
from ..event_utls.producer import Producer
from ..utils.utils import generate_unique_id

_logger = init_logger(__name__)


class MessageBusWriter:
    def __init__(self, service_type, simulations=[], is_sync=False, producer=None):
        self.producer = Producer() if producer is None else producer
        self.cmd_queue = []
        self.service_type = service_type
        self.simulations = simulations
        self._is_sync = is_sync
        self.send_messages = []

    def reset(self):
        self.send_messages = []

    def commit(self):
        data = SetMachineMessage(is_last_message=False, write_commands=self.cmd_queue)
        package_id = generate_unique_id()
        self.producer.sync_produce(PROPOSAL_TOPIC, data.serialize(), Headers(package_id=package_id, source=self.service_type, msg_type=self.service_type))
        self.cmd_queue = []
        self.send_messages.append(package_id)

        if self._is_sync:
            self.wait_for(package_id)

    def __call__(self, channel, device, _property, commit=True, **kwargs):
        if commit:
            self.cmd_queue.append(WriteCommand(channel=channel, device_name=device, property=_property, params=kwargs))
            return

        data = SetMachineMessage(is_last_message=False, write_commands=[WriteCommand(channel=channel, device_name=device, property=_property, params=kwargs)])
        package_id = generate_unique_id()
        self.producer.sync_produce(PROPOSAL_TOPIC, data.serialize(), Headers(package_id=package_id, source=self.service_type, msg_type=self.service_type))
        self.send_messages.append(package_id)
        if self._is_sync:
            self.wait_for(package_id)
        return package_id

    def wait_for(self, package_id):
        if len(self.simulations) > 0:
            for sim in self.simulations:
                res = self._wait_for_write(package_id, sim)
                if res is None:
                    raise TimeoutError(f"timeout reached for package_id={package_id} and service_type={sim}.")
        else:
            res = self._wait_for_write(package_id, MachineService.__name__)
            if res is None:
                raise TimeoutError(f"timeout reached for package_id={package_id} and service_type={MachineService.__name__}.")

    def _wait_for_write(self, package_id, service_type, timeout=30, poll_time=0.05):
        res = requests.get(f'http://{OBSERVER_URL}/find/{package_id}/{service_type}')
        if res.status_code == 200:
            return res.json()
        time_counter = 0.0
        while(res.status_code == 202):
            time.sleep(poll_time)
            time_counter += poll_time
            res = requests.get(f'http://{OBSERVER_URL}/find/{package_id}/{service_type}')
            if res.status_code == 200:
                _logger.debug(f"service name = {service_type}")
                return res.json()
            if time_counter >= timeout:
                _logger.error(f"timeout reached for package_id={package_id} and service_type={service_type}.")
                return None
            poll_time *= 2.0



import threading


class TimerThread(threading.Thread):
    def __init__(self, time: float, func):
        """_summary_
        An thread which will call func periodically. 
        :param time: Defines the time period in seconds.
        :type time: float
        :param func: function that will be called after time is passed
        :type func: _type_
        """
        super().__init__()
        self._stop_event = threading.Event()
        self.time = time
        self.func = func

    def stopped(self):
        return self._stop_event.is_set()

    def stop(self):
        self._stop_event.set()

    def run(self):
        while not self._stop_event.wait(self.time):
            self.func()



def load_at_optic(optic: str):
    if optic == 'v24':
        from .physics_based.p3_elements_v24 import ring
    elif optic == 'v24_c4l':
        from .physics_based.p3_elements_v24_c4l import ring
    else:
        raise ValueError(f"Optic {optic} is unknown.")
    return ring



from ctypes import Union
import dataclasses
from typing import List, Union

from .message import Message


@dataclasses.dataclass
class AgentResultMessage(Message):
    '''
    Result of the Agent
    '''
    result: Union[str, dict] # can be an url [string] to the agents database or a dict with the results. 
    error_message: str
    send_messages: List[str]
    error_code: int


from time import time


def check_if_sql_db_table_exists(connection, tablename):
    cursor = connection.cursor()

    sql_query = f"select COUNT(*) from information_schema.tables where TABLE_NAME='{tablename}'"
    cursor.execute(sql_query)
    table_count = cursor.fetchone()[0]
    connection.commit()
    if table_count == 1:
        cursor.close()
        return True
    cursor.close()
    return False


def check_if_sql_table_is_filled(connection, tablename):
    cursor = connection.cursor()

    sql_query = f"select COUNT(*) from {tablename}"
    cursor.execute(sql_query)
    count = cursor.fetchone()[0]
    connection.commit()
    if count > 0:
        cursor.close()
        return True
    cursor.close()
    return False



from typing import List, Dict


from ..service import Service
from ..logger import init_logger
from ..topics import SIMULATION_EVENTS
from ..messages.message import Headers
from ..messages.update_message_types import UpdateMessage
from src.acss_core.event_utls.consumer import consume
from ..utils.TimerThread import TimerThread

_logger = init_logger(__name__)


class TimedSimService(Service):
    def __init__(self, name, write, read, time: int):
        super().__init__(name)
        self.category = 'Simulation'
        self.read = read

        self._write = write
        self._updated_channels = []

        self.updated_devices: Dict[int, List[List[str]]] = {}
        self._thread = TimerThread(time, self.simulate)

    def write(self, channel: str, device: str, _property, **kwargs):
        self._updated_channels.append(channel)
        self._write(channel, device, _property, **kwargs)

    def stop(self):
        super().stop()
        self._thread.stop()
        self._thread.join()

    def _simulate(self):
        self.simulate()
        self.sim_updated_event(package_id='None',
                               msg=UpdateMessage(source=self.name, updated=self._updated_channels))

    def simulate(self):
        pass

    def post_init(self):
        self._thread.start()
        self._thread.join()

    def sim_updated_event(self, package_id, msg):
        self.producer.sync_produce(SIMULATION_EVENTS, msg.serialize(), Headers(package_id=package_id, source=self.type))



import json
import dataclasses
import typing
import time
import datetime
import numpy as np
from io import BytesIO

import dacite
import confluent_kafka

from ..utils.utils import generate_unique_id


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        elif isinstance(o, np.ndarray):
            memfile = BytesIO()
            np.save(memfile, o)
            memfile.seek(0)
            serialized = json.dumps(memfile.read().decode('latin-1'))
            return serialized
        return super().default(o)


class Headers:
    MSG_TYPE_REGISTER = "register"
    MSG_TYPE_DEREGISTER = "deregister"
    MSG_TYPE_INFO = "info"

    def __init__(self,  source: str, package_id=None,  only_for: typing.List[str] = None, msg_type=None,
                 execution_loc=None):
        self.package_id = generate_unique_id() if package_id == None else package_id
        self.source = source
        self.only_for = only_for
        self.msg_type = msg_type  # used by the communication between scheduler and services
        self.execution_loc = execution_loc

    def to_kafka_headers(self) -> typing.Dict[str, bytes]:
        headers = {'source': bytes(self.source, encoding='utf-8')}
        if self.package_id:
            headers['package_id'] = bytes(self.package_id, encoding='utf-8')
        if self.only_for:
            headers['only_for'] = bytes(','.join(self.only_for), encoding='utf-8')
        if self.msg_type:
            headers['msg_type'] = bytes(self.msg_type, encoding='utf-8')
        if self.execution_loc:
            headers['execution_loc'] = bytes(self.execution_loc, encoding='utf-8')

        return headers

    def is_source(self, service_name):
        if self.source:
            return self.source == service_name
        return False

    def is_message_for(self, service_name):
        if self.only_for:
            return service_name in self.only_for
        return False

    def __str__(self):
        return f'Header:[ source: {self.source}, package_id: {self.package_id}, only_for: {",".join(self.only_for) if self.only_for else "None"}, ' \
               f'msg_type: {self.msg_type if self.msg_type else "None"}, execution_log: {self.execution_loc} ]'

    @classmethod
    def from_kafka_headers(cls, headers: typing.List[typing.Tuple[str, bytes]]) -> 'Headers':
        param = {}
        for key, value in headers:
            if key == "only_for":
                param[key] = value.decode('utf-8').split(',') if b',' else [value.decode('utf-8')]
            else:
                param[key] = value.decode('utf-8')
        return cls(**param)


@dataclasses.dataclass
class Message:
    def serialize(self) -> str:
        return json.dumps(self, cls=EnhancedJSONEncoder)

    @classmethod
    def deserialize(cls, messages: typing.Union[typing.List[confluent_kafka.Message], typing.List[bytes]]) -> 'Message':
        """
        Deserialize the message to the specific message data class.
        @param messages: Message received by the Kafa which type is List[confluent_kafka.Message] or by the http client
        which type is List[bytes]
        @return: A dataclass from the specific message
        """
        cls = Message._deserialize(class_type=cls, messages=messages)
        return cls

    @classmethod
    def from_byte_string(cls, byte_str):
        message_dict = json.loads(byte_str)
        return Message._from_dict_to_dataclass(cls, message_dict)


    @staticmethod
    def _deserialize_ndarray(serialized_array):
        memfile = BytesIO()
        memfile.write(json.loads(serialized_array).encode('latin-1'))
        memfile.seek(0)
        return np.load(memfile)

    @staticmethod
    def _deserialize_dict(serialized_dict):
        for key, value in serialized_dict.items():
            # is value a numpy array decoded string?
            if isinstance(value, str) and value[:12] == r'"\u0093NUMPY':
                serialized_dict[key] = Message._deserialize_ndarray(value)
        return serialized_dict

    @staticmethod
    def _from_dict_to_dataclass(class_type: 'Message', message_dict):
        return dacite.from_dict(data_class=class_type, data=message_dict,
                                config=dacite.Config(type_hooks={np.ndarray: Message._deserialize_ndarray,
                                                                 dict: Message._deserialize_dict}))

    @staticmethod
    def _deserialize(class_type: 'Message',
                     messages: typing.Union[typing.List[confluent_kafka.Message], typing.List[bytes]]):

        merged_message_values = ''
        for message in messages:
            merged_message_values += message.decode('utf-8') if type(message) == bytes else message.value().decode(
                'utf-8')

        return class_type.from_byte_string(merged_message_values)

    def get_readable_timestamp(self, fmt: str = '%Y-%m-%d %H:%M:%S') -> str:
        return datetime.datetime.strftime(time.localtime(self.timestamp), fmt)


@dataclasses.dataclass
class AliveMessage(Message):
    info: str
    type: str
    status: str
    category: str


@dataclasses.dataclass
class ControlMessage(Message):
    name: str
    command: str



import json
import typing
import time
import requests
from abc import ABC

from confluent_kafka import TIMESTAMP_NOT_AVAILABLE

from .config import KAFKA_SERVER_URL, REGISTER_URL
from .logger import init_logger
from .utils.utils import wait_until_server_is_online
from .event_utls.consumer_thread import ConsumerThread
from .event_utls.producer import Producer
from .messages.message import Headers
from .topics import CONTROL_TOPIC, MACHINE_EVENTS
from .topics import IS_ALIVE
from .utils.TimerThread import TimerThread
from .messages.message import Headers, AliveMessage

_logger = init_logger(__name__)


class Service(ABC):
    def __init__(self, name, **kwargs):
        self.name: str = name
        self.creation_timestamp = int(
            time.time() * 1000)  # The creation_timestamp is the number of milliseconds since the epoch (UTC).
        self.control_log_consumer = None
        self.producer = None
        self.type = self.__class__.__name__
        self.category = ''
        self.status = "INIT"

    def to_json(self):
        return json.dumps({'name': self.name})

    @classmethod
    def from_json(cls, serialized_service_obj):
        service_obj_json = json.loads(serialized_service_obj)
        name = service_obj_json.get('name')
        if not name:
            raise ValueError('Key name is not in json object.')
        return cls(name)

    def _control_log_handler(self, msg):
        _logger.debug(f"[{self.name}] call _control_log_handler")
        timestamp_type, timestamp = msg.timestamp()
        if timestamp_type == TIMESTAMP_NOT_AVAILABLE:
            _logger.debug(
                f"[{self.name}] receive a message without a timestamp")
            return
        if timestamp < self.creation_timestamp:
            return
        message = json.loads(msg.value().decode('utf-8'))
        _logger.debug(
            f"[{self.name}] call _control_log_handler receive message: {message}")
        command = message.get('command')
        if self.name == message.get('name'):
            if command == "start":
                self.start()
            elif command == "stop":
                self.stop()

    def info(self) -> str:
        return "No info string is defined. The concrete service have to overload info()."

    def _init_alive_thread(self):
        self.producer.wait_until_topics_created([IS_ALIVE], poll_time=1, max_iterations=30)
        self.alive_thread = TimerThread(0.1, self._alive)
        self.alive_thread.start()

    def post_init(self):
        pass

    def _alive(self):
        self.producer.async_produce(topic=IS_ALIVE, value=AliveMessage(info=self.info(), type=self.type, status=self.status, category=self.category).serialize(), headers=Headers(self.name))

    def _register(self):
        wait_until_server_is_online(REGISTER_URL, logger=_logger)
        res = requests.get(f"http://{REGISTER_URL}/register/{self.name}")
        _logger.debug(f"receive status code {res.status_code}")
        _logger.debug(f"error: {res.content}")
        if res.status_code != 200:
            raise Exception(f"Registration of service failed. An service with the name {self.name} is already running/")
        self.producer = Producer()

    def init_local(self):
        _logger.debug(f"[{self.name}] call init_local")

        self._register()

        self.producer = Producer()
        wait_until_server_is_online(url=KAFKA_SERVER_URL, logger=_logger)
        # wait maximal ~30s to let topics be created. Just for startup
        self.producer.wait_until_topics_created([CONTROL_TOPIC, MACHINE_EVENTS], poll_time=1, max_iterations=30)

        self.control_log_consumer = ConsumerThread(topics=[CONTROL_TOPIC], group_id=self.name + "_" + CONTROL_TOPIC,
                                                   bootstrap_server=KAFKA_SERVER_URL,
                                                   message_handler=self._control_log_handler, consumer_started_hook=None)

        self.control_log_consumer.start()
        self.control_log_consumer.wait_until_ready()

        self._init_alive_thread()

        self.status = 'RUNNING'
        self.post_init()
        _logger.debug(f"[{self.name}] exit init_local")

        # wait until stop() is executed.
        self.control_log_consumer.join()

    def stop(self):
        _logger.debug(f"stop is called on service {self.name}")
        if self.alive_thread:
            self.alive_thread.stop()
            self.alive_thread.join()

        # note: .join() is already called in init_local which is called mainThread
        self.control_log_consumer.stop()

    def send(self, topic, message, headers):
        _logger.debug(f"[{self.name}] call send {str(headers)}")
        self.producer.sync_produce(topic=topic,
                                   value=message.serialize(),
                                   headers=headers)

    # used for pickle class
    def __getstate__(self):
        odict = {'name': self.name}
        return odict

    # used for pickle class
    def __setstate__(self, state):
        kwargs = {}
        self.__dict__ = type(self)(state['name'], **kwargs).__dict__



import time
from ..logger import init_logger
from ..utils.sql_utils import check_if_sql_db_table_exists, check_if_sql_table_is_filled
import mysql.connector
_logger = init_logger(__name__)


class SqlDBConnector():
    def __init__(self, user, pw, host, port, database):
        self.sql_con = self.connect_to_db(user, pw, host, port, database)

    def check_if_sql_db_table_exists(self, tablename):
        return check_if_sql_db_table_exists(self.sql_con, tablename)

    def check_if_sql_table_is_filled(self, tablename):
        return check_if_sql_table_is_filled(self.sql_con, tablename)

    def connect_to_db(self, user, pw, host, port, database):
        while(True):
            try:
                cnx = mysql.connector.connect(user=user, password=pw, host=host, port=port, database=database)
                break
            except Exception as e:
                if e.errno == mysql.connector.errorcode.CR_UNKNOWN_HOST or e.errno == mysql.connector.errorcode.CR_CONN_HOST_ERROR:
                    time.sleep(0.5)
                    _logger.debug(f"Try to connect again. error: {e}")
                    continue
                else:
                    raise e
        return cnx

    def wait_until_table_is_created(self, table: str):
        while(True):
            try:                
                if self.check_if_sql_db_table_exists(table) and self.check_if_sql_table_is_filled(table):
                    return
            except mysql.connector.errors.ProgrammingError as mysql_error:
                if mysql_error.errno == mysql.connector.errorcode.ER_NO_SUCH_TABLE:                
                    _logger.warning(f"Table {table} doesn't exists.")
                    time.sleep(0.5)


import dataclasses
from typing import List
from xmlrpc.client import Boolean

from .message import Message


@dataclasses.dataclass
class UpdateMessage(Message):
    source: str
    updated: List[str]


@dataclasses.dataclass
class WriteCommand():
    channel: str
    device_name: str
    property: str
    params: dict


@dataclasses.dataclass
class SetMachineMessage(Message):
    is_last_message: Boolean
    write_commands: List[WriteCommand]



from typing import List
from .kafka_client import KafkaPipeClient, ServiceId
from ..messages.agent_result_message import AgentResultMessage


class ServiceClient():
    def __init__(self, pipe_client: KafkaPipeClient, names: List[str], type: str,  info: str) -> None:
        self.pipe = pipe_client
        self.info = "No info test." if info == None else info
        self.type = type
        self.names = names

    def stop(self, names=None):
        supspending_services = self.names if names is None else names
        for name in supspending_services:
            self.pipe.stop_service(name)


class MachineClient(ServiceClient):
    def __init__(self, names: List[str], pipe_client: KafkaPipeClient, type: str,  info: str) -> None:
        super().__init__(pipe_client, names, type, info)

    def is_processed(self, service_id: ServiceId):
        return False if self.pipe.wait_for_simulation(service_id, self.type) is None else True


class AgentClient(ServiceClient):
    def __init__(self, pipe_client: KafkaPipeClient, names: List[str], type: str,  info: str) -> None:
        super().__init__(pipe_client, names, type, info)
        self._service_id = None

    def reconfig(self, config, sync=False):
        p_id = self.pipe.reconfig(names=[self.type], config_data=config)
        if sync:
            return self.pipe.get_service_results(ServiceId(p_id, self.type))

    def run(self, params=None, wait_for_sims=[]):
        self._service_id = self.pipe.run_service(self.type, params)
        if len(wait_for_sims) > 0:
            agt_res = self.pipe.get_service_results(self._service_id)
            ids = agt_res[0].send_messages
            for sim in wait_for_sims:
                agt_result = self.pipe._get_results_from_observer(ids, sim, timeout=30, poll_time=0.05)
                if agt_result is None:
                    raise TimeoutError()
            return None
        return self._service_id

    def get_result(self):
        return self.pipe.get_service_results(self._service_id)[0]



import os
import subprocess
from enum import Enum
from typing import Optional

from src.acss_core.logger import init_logger
from src.acss_core.config import ACSS_CONFIG_FILEPATH

PROFILE = "pipeline"

ACSS_CONFIG_FILEPATH_HOST = os.environ.get("ACSS_CONFIG_FILEPATH_HOST")

_logger = init_logger(__name__)


class DOCKER_COMPOSE_FILTER_BY_STATUS(Enum):
    RUNNING = 'running'
    RESTARTING = 'restarting'
    PAUSED = 'paused'
    STOPPED = 'stopped'


class DockerComposeClient():
    def __init__(self, max_num_of_container: int = 64) -> None:
        self.max_num_of_container = max_num_of_container
        self.host_config_path = os.environ.get("HOST_CONFIG_PATH")

    def get_services(self, filter: Optional[DOCKER_COMPOSE_FILTER_BY_STATUS] = None):
        filter_option = f"--filter status={filter.value}" if filter != None else ""
        ps_str = subprocess.check_output(f"docker-compose -p {PROFILE} ps {filter_option} --services".split(), timeout=10).decode('utf-8').strip()
        split = ps_str.split('\n')
        return split

    def run_service(self, name: str, service_type: str, mode: str):
        running_container = DockerComposeClient.get_services(DOCKER_COMPOSE_FILTER_BY_STATUS.RUNNING)
        if len(running_container) == self.max_num_of_container:  # no free service where script can run
            _logger.warning(f"Maximum running service containers of {self.max_num_of_services} reached. It is not possible to start more service container.")
            return False
        #subprocess.run(args=f'docker-compose -p {PROFILE} run --rm --no-deps -e SERVICE_NAME={name} -e SERVICE_MODE={mode} {service_type}'.split())
        command = f'docker-compose -p {PROFILE} run --rm --no-deps -e SERVICE_NAME={name} -e SERVICE_MODE={mode} -e ACSS_CONFIG_FILEPATH={ACSS_CONFIG_FILEPATH} -v {ACSS_CONFIG_FILEPATH_HOST}:{ACSS_CONFIG_FILEPATH} {service_type}'
        print(command)
        subprocess.Popen(command.split())
        return True


if __name__ == "__main__":
    compose_client = DockerComposeClient()
    print(DockerComposeClient.get_services(filter=DOCKER_COMPOSE_FILTER_BY_STATUS.RUNNING))
    print(f"active service containers: {compose_client.active_service_container} max num of services: {compose_client.max_num_of_services}")
    compose_client.start_service_container(num=4)
    print(f"active service containers: {compose_client.active_service_container} max num of services: {compose_client.max_num_of_services}")



import sys
from typing import Union, List

from confluent_kafka import Consumer, KafkaError, KafkaException, TopicPartition

from ..config import KAFKA_SERVER_URL


def consume(topics: Union[List[str], List[TopicPartition]], group_id: str, timeout_poll: int = 0.1, max_session: int = 0,
            consumer_started_hook=None, all_msg_hook=None, stop_hook=None, bootstrap_server=KAFKA_SERVER_URL):
    """        
    Returns a message, when a new message is available in subscribed topics.
    @param topics: Topics that will be consumed. Can be a List of strings with the topic names or a list of TopicPartitions
    @param group_id: Consumer's group_id
    @param timeout_poll: Maximal waiting time for a new message.
    @param max_session: Maximum iterations, when the maximum is reached a TimeoutError exception will be thrown.
    @param consumer_started_hook: A function that is called after the consumer is initialized and is conusming the topic.
    @param all_msg_hook: A function that is called after the consumer.poll function.
    @param stop_hook: A function that returns bool. If the functions returns true, consuming will be stopped.
    @return: Message of type KafkaMessage.        
    """
    settings = {
        'bootstrap.servers': bootstrap_server,
        'group.id': group_id,
        'auto.offset.reset': 'latest',
        'enable.partition.eof': True if consumer_started_hook else False
    }

    print(f"consumer settings: {settings}")

    consumer = Consumer(settings)

    if all(isinstance(x, str) for x in topics):
        consumer.subscribe(topics)
    else:
        consumer.assign(topics)
    session = 0
    message_published = False
    while True:
        if stop_hook and stop_hook():
            break
        msg = consumer.poll(timeout_poll)
        if all_msg_hook:
            all_msg_hook(msg)
        if max_session and session > max_session:
            raise TimeoutError
        if msg is None:
            session += 1
            continue
        if msg.error():
            if msg.error().code() == KafkaError._PARTITION_EOF:
                # https://github.com/confluentinc/confluent-kafka-python/issues/156
                # https://github.com/confluentinc/confluent-kafka-python/issues/581
                if not message_published and consumer_started_hook:
                    consumer_started_hook()
                    message_published = True
                # End of partition event
                sys.stderr.write('%% %s [%d] reached end at offset %d\n' %
                                 (msg.topic(), msg.partition(), msg.offset()))
            elif msg.error():
                # https://github.com/edenhill/librdkafka/issues/1987
                if msg.error().code() == KafkaError._TRANSPORT:
                    sys.stderr.write(
                        'GroupCoordinator response error: Local: Broker transport failure')
                else:
                    raise KafkaException(msg.error())
        yield msg
    consumer.close()



import time
import typing

import confluent_kafka

from ..config import KAFKA_SERVER_URL
from ..messages.message import Headers
from ..topics import NUM_OF_PARTITION


class Producer:
    def __init__(self, event_server_url=KAFKA_SERVER_URL):
        self._producer = confluent_kafka.Producer({'bootstrap.servers': event_server_url,
                                                   'topic.metadata.refresh.interval.ms': 10,
                                                   'message.max.bytes': 52428800})  # Because of slow on first message. 1.7.0 should fix it but its still there. https://github.com/confluentinc/confluent-kafka-dotnet/issues/701
        #self._producer = confluent_kafka.Producer({'bootstrap.servers': event_server_url, 'message.max.bytes': 52428800})

    def _produce(self, topic, value, headers: Headers, is_async=False):

        self._producer.produce(topic, value=value, headers=headers.to_kafka_headers())
        if is_async:
            self._producer.poll(1)
        else:
            self._producer.flush()

    def async_produce(self, topic, value, headers):
        self._produce(topic, value, headers=headers, is_async=True)

    def all_partitions_produce(self, topic, value, headers):
        for p_id in range(NUM_OF_PARTITION):
            self._producer.produce(topic, partition=p_id, value=value, headers=headers.to_kafka_headers())
        self._producer.poll(1)

    def sync_produce(self, topic, value, headers):
        self._produce(topic, value, headers=headers, is_async=False)

    def get_existing_topics(self) -> typing.List[str]:
        metadata = self._producer.list_topics()
        self._producer.flush()
        return [key for key, _ in metadata.topics.items()]

    def is_topic_created(self, topic):
        metadata = self._producer.list_topics(topic=topic)
        self._producer.flush()
        # there is just one topic in topics
        topic_metadata = metadata.topics.get(topic)
        if not topic_metadata:
            return False
        else:
            return metadata.topics[topic].error

    def get_search_for_topic(self, topic) -> typing.List[str]:
        metadata = self._producer.list_topics(topic=topic)
        self._producer.flush()

        res = {key: value.error for key, value in metadata.topics.items()}
        for topic, error in res.items():
            print(f"Error messages from topic '{topic}': {error}")
        return res

    def wait_until_topics_created(self, topics, poll_time=0.5, max_iterations=0):
        # Wait for full metadata state for topics
        # https://github.com/confluentinc/confluent-kafka-python/issues/524
        all_topics_created = False
        iter_count = 0
        while not all_topics_created and iter_count <= max_iterations:
            all_topics_created = True
            for topic in topics:
                existing_topics = self.get_search_for_topic(topic=topic)
                if existing_topics.get(topic) and existing_topics[topic]:
                    all_topics_created = False
                    time.sleep(poll_time)
                    break
            if max_iterations != 0:
                iter_count += 1

        if max_iterations < iter_count:
            raise TimeoutError



# num of kafka partitions
NUM_OF_PARTITION = 3

SERVICE_INPUT_TOPIC = 'service-input-event'
SERVICE_OUTPUT_TOPIC = 'service-output-event'
SERVICE_OBSERVE_TOPIC = 'service-observed-event'
RECONFIG_TOPIC = 'reconfig-event'
CONTROL_TOPIC = "control-event"
IS_ALIVE = "healthy"
ACTIVE_SERVICES_UPDATED_TOPIC = "active-services-updated-event"
AGENT_EVENTS = "agent-result-event"
# machine topics
MACHINE_EVENTS = 'machine-set-event'
SIMULATION_EVENTS = 'simulation-event'
PROPOSAL_TOPIC = 'proposal-event'



from typing import Tuple, Dict
from .kafka_client import KafkaPipeClient
from .service_client import AgentClient, MachineClient


def get_services() -> Tuple[Dict[str, AgentClient], Dict[str, MachineClient]]:
    pipe = KafkaPipeClient()
    agents = {}
    machines = {}

    services_sorted_by_type = {}
    running_services = pipe.get_running_service_info()
    # Sort by service type
    for name, info in running_services.items():
        s_type = info['type']
        found_s_type = services_sorted_by_type.get(s_type)
        if found_s_type is not None:
            found_s_type.append(name)
        else:
            services_sorted_by_type[s_type] = [name]
    for s_type, names in services_sorted_by_type.items():

        first_service = running_services[names[0]]

        if first_service['cat'] == 'Simulation':
            machines[s_type] = MachineClient(pipe_client=pipe, names=names, info=first_service['info'], type=first_service['type'])
        else:
            agents[s_type] = AgentClient(pipe_client=pipe, names=names, info=first_service['info'], type=first_service['type'])
    return agents, machines





import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="acss_core",
    version="0.0.1",
    author="Michael Boese",
    author_email="michael.boese@desy.de",
    description="Pipeline for accelerators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/desy-ml/acss-core",
    project_urls={
        "Bug Tracker": "https://github.com/desy-ml/acss-core/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    install_requires=['confluent-kafka', 'requests', 'pyyaml', 'python-logstash', 'numpy', 'dacite', 'mysql-connector-python', 'accelerator-toolbox'],
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
)



from logging import log
import time
import datetime
import socket
import uuid


def create_timestamp() -> int:
    """
    Creates a timestamp.
    @return: number of milliseconds since the epoch (UTC).
    """
    return datetime.datetime.utcnow().timestamp() * 1000


def timestamp_to_date(t: int) -> str:
    return str(datetime.datetime.utcfromtimestamp(t / 1000.0))


def check_connection(url: str, logger=None):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(20)
    result = sock.connect_ex((url.split(':')[0], int(url.split(':')[1])))
    if result != 0:
        if logger != None:
            logger.error(f"Can't connect to address: {url}")
        raise ConnectionError(f"Can't connect to address: {url}")
    sock.close()


def wait_until_server_is_online(url, logger, sleep_time: int = 2, retries: int = 15) -> None:
    for _ in range(retries):
        try:
            check_connection(url, logger)
        except ConnectionError as e:
            time.sleep(sleep_time)
            if logger != None:
                logger.info(f"Try again to reconnect to server {url}")


def generate_unique_id():
    return str(uuid.uuid4())



import logging
import sys
import os
import logstash

from .utils.utils import wait_until_server_is_online
from .config import LOGSTASH_SERVER_URL, LOGSTASH_PORT


def init_logger(logger_name, log_lvl=logging.DEBUG):
    _logger = logging.getLogger(logger_name)
    _logger.setLevel(log_lvl)
    if LOGSTASH_SERVER_URL != None and LOGSTASH_PORT != None:
        wait_until_server_is_online(LOGSTASH_SERVER_URL + ":" + LOGSTASH_PORT, logger=_logger)
        _logger.addHandler(logstash.TCPLogstashHandler(LOGSTASH_SERVER_URL, LOGSTASH_PORT, version=1))
        _logger.info(f"Logging is set to logstash")
    else:
        _logger.warning(f"Environment value LOGSTASH_SERVER_URL isn't set. Service will not log in central Log DB only to stdout.")
        _logger.info("Logging is set to stdout.")

    return _logger


path_to_logfile = os.environ.get('MPY_ML_PIPE_KAFKA_LOG_FILE_PATH')

if path_to_logfile:
    logging.basicConfig(
        filename=path_to_logfile,
        format='[%(asctime)s.%(msecs)03d] %(levelname)-4s  %(message)s',
        level=logging.INFO,
        datefmt='%Y-%m-%d %H:%M:%S')
else:
    logging.basicConfig(
        stream=sys.stdout,
        format='[%(asctime)s.%(msecs)03d] %(levelname)-4s  %(message)s',
        level=logging.INFO,
        datefmt='%Y-%m-%d %H:%M:%S')



import json
from abc import abstractmethod
from typing import Optional, Tuple
import requests
import time

from confluent_kafka import TIMESTAMP_NOT_AVAILABLE

from .logger import init_logger
from .config import KAFKA_SERVER_URL
from .service import Service
from .event_utls.consumer import consume
from .messages.message import Headers
from .messages.simple_service_message import SimpleServiceMessage
from .topics import AGENT_EVENTS, RECONFIG_TOPIC, SERVICE_INPUT_TOPIC
from .messages.update_message_types import *
from .messages.agent_result_message import AgentResultMessage
from .adapter.MsgBusWriter import MessageBusWriter

_logger = init_logger(__name__)


class SimpleService(Service):
    def __init__(self, name, read, is_sync=False, simulations=[]):
        super().__init__(name)
        self.write = MessageBusWriter(self.type, simulations, is_sync=is_sync)
        self.read_hook = read
        self.current_package_id = None
        self.simulations = simulations
        self.__stop_consumer_loop = False

    @abstractmethod
    def proposal(self, params: Optional[dict]) -> Optional[Tuple[dict, int, str]]:
        """[summary]
        Function that will be called when the Agent is triggered.
        :param params: Optional parameter for configuration or setting internal parameter
        :type params: Optional[dict]
        :return: A tuple consisting of a dictionary which is containing the result of the Agent that will be stored,
        an error code and and error message. All parameter are optional and can be set to None if not needed.
        :rtype: Optional[Tuple[dict, int, str]]
        """
        pass

    def read(self, channel, device, _property, **kwargs):
        if self.read_hook is None:
            raise KeyError("read_hook isn't set.")
        return self.read_hook(channel, device, _property, **kwargs)

    def reconfig_event(self, msg):
        """[summary]
        Function that can be overloaded by the concret service to reconfig without restarting
        :param msg: [description]
        :type msg: [type]
        """
        pass

    def _does_local_storage_exists(self):
        # ToDo: Not implemented yet
        return False

    def stop(self):
        self.__stop_consumer_loop = True
        super().stop()
        _logger.debug(f"exit stop on service {self.name}")

    def __consumer_stop_hook(self):
        return self.__stop_consumer_loop

    def post_init(self):
        for msg in consume([SERVICE_INPUT_TOPIC, RECONFIG_TOPIC], stop_hook=self.__consumer_stop_hook, timeout_poll=0.1, group_id=self.type, bootstrap_server=KAFKA_SERVER_URL):
            self.service_input_handler(msg)

        _logger.debug("end consuming")

    def _service_input_event_handler(self, msg):
        data = SimpleServiceMessage.deserialize([msg])
        result = self.proposal(data.params)
        # store message
        agent_result, error_code, error_message = result if result != None else ({}, 0, 'None')
        if self._does_local_storage_exists():
            # TODO: store the agent_result in local db and set agent_result to db url
            pass
        self.producer.async_produce(AGENT_EVENTS, AgentResultMessage(agent_result, error_message, self.write.send_messages, error_code).serialize(),
                                    Headers(package_id=self.current_package_id, source=self.type, msg_type=self.type))
        self.write.reset()

        _logger.debug(f"send result message to topic '{AGENT_EVENTS}'")

    def service_input_handler(self, msg, **kwargs):
        timestamp_type, timestamp = msg.timestamp()
        if timestamp_type == TIMESTAMP_NOT_AVAILABLE:
            _logger.debug(f"[{self.name}] receive a message without a timestamp")
            return
        headers = Headers.from_kafka_headers(msg.headers())
        self.current_package_id = headers.package_id
        #self.machine_adapter.active_package_id = headers.package_id
        _logger.debug(f'[{self.name}] call service_input_handler receive headers: {str(headers)}')
        _logger.debug(f"Received message from topic '{msg.topic()}'")

        if headers.is_message_for(self.type) or headers.is_message_for(self.name):
            if msg.topic() == SERVICE_INPUT_TOPIC:
                self._service_input_event_handler(msg)

            elif msg.topic() == RECONFIG_TOPIC:
                _logger.debug(f"reconfig message received")
                config_data = json.loads(msg.value().decode('utf-8'))
                self.reconfig_event(config_data)
                self.producer.async_produce(AGENT_EVENTS, AgentResultMessage(config_data, "None", [self.current_package_id], 0).serialize(),
                                            Headers(package_id=self.current_package_id, source=self.type, msg_type=self.type))
                return

        _logger.debug(f'[{self.name}] end service_input_handler')
